package filegateway

import (
	"encoding/json"
	"testing"
	"text/template"
	"time"

	"github.com/stretchr/testify/assert"
	stdlib "orus.io/xbus/xbus-stdlib/lib"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/testutils"
)

func must(tb testing.TB, err error) {
	if err != nil {
		tb.Fatal(err)
	}
}

func newTestOutputs() *testOutputs {
	return &testOutputs{}
}

type testOutputs struct {
	Outputs []*xbus.EnvelopeReceiver
}

func (t *testOutputs) OpenOutput(
	context *api.ProcessingContext, output string, eventTypes ...string,
) *xbus.EnvelopeSender {
	receiver := xbus.NewEnvelopeReceiverEmpty()
	t.Outputs = append(t.Outputs, receiver)
	sender := xbus.NewEnvelopeSender(
		eventTypes, 1024*512,
		func(e *api.Envelope) (api.EnvelopeAck_ReceptionStatus, error) {
			receiver.AddFragment(*e)
			return receiver.Status(), receiver.LastError()
		})
	return sender
}

func TestEmitter(t *testing.T) {
	t.Run("basics", func(t *testing.T) {
		var (
			fname1 = t.Name() + "/f1.txt"
			fname2 = t.Name() + "/f2.log"
			fname3 = t.Name() + "/f3.txt"
		)
		fs := getMemoryFilestem(".")
		fs.pollingInterval = 5 * time.Millisecond
		logger := testutils.GetTestLogger(t)
		outputs := newTestOutputs()

		emitter := Emitter{
			openOutput: outputs.OpenOutput,
			log:        logger,
			fs:         fs,
			pattern:    t.Name() + "/*.txt",
			msgType:    "test",
			postAction: "delete",
		}

		must(t, emitter.Startup())

		time.Sleep(10 * time.Millisecond)
		must(t, writeFile(fs, fname1, "f1 content"))

		time.Sleep(10 * time.Millisecond)
		must(t, writeFile(fs, fname2, "f2 content"))

		time.Sleep(10 * time.Millisecond)
		must(t, writeFile(fs, fname3, "f3 content"))

		time.Sleep(20 * time.Millisecond)

		must(t, emitter.Shutdown())

		if l := len(outputs.Outputs); l != 2 {
			t.Fatalf("Expected 2 emissions, got %d", l)
		}

		env1, err := outputs.Outputs[0].Complete(time.Second)
		must(t, err)
		env2, err := outputs.Outputs[1].Complete(time.Second)
		must(t, err)
		var m1, m2 stdlib.FileMeta
		must(t, json.Unmarshal(env1.Events[0].Items[0], &m1))
		must(t, json.Unmarshal(env2.Events[0].Items[0], &m2))
		assert.Equal(t, "f1.txt", m1.Filename)
		assert.Equal(t, "f1 content", string(env1.Events[0].Items[1]))
		assert.Equal(t, "f3.txt", m2.Filename)
		assert.Equal(t, "f3 content", string(env2.Events[0].Items[1]))
		assert.NotContains(t, fs.Files, fname1)
		assert.Contains(t, fs.Files, fname2)
		assert.NotContains(t, fs.Files, fname3)
	})

	t.Run("post-action 'rename'", func(t *testing.T) {
		var (
			fname1 = t.Name() + "/f1.txt"
		)
		fs := NewMemoryFilesystem(".")
		fs.pollingInterval = 5 * time.Millisecond
		logger := testutils.GetTestLogger(t)
		outputs := newTestOutputs()

		emitter := Emitter{
			openOutput: outputs.OpenOutput,
			log:        logger,
			fs:         fs,
			pattern:    t.Name() + "/*.txt",
			msgType:    "test",
			postAction: "rename",
			renameTemplate: template.Must(template.New("rename").Parse(
				"{{ .Dir }}/sub/{{ .WithoutExt }}-{{ .EnvelopeID }}-post{{ .Ext }}")),
		}

		must(t, emitter.Startup())

		must(t, writeFile(fs, fname1, "f1 content"))

		time.Sleep(20 * time.Millisecond)

		must(t, emitter.Shutdown())

		if l := len(outputs.Outputs); l != 1 {
			t.Errorf("Expected 1 emissions, got %d", l)
			if l < 1 {
				t.FailNow()
			}
		}

		env1, err := outputs.Outputs[0].Complete(time.Second)
		must(t, err)
		var m1 stdlib.FileMeta
		must(t, json.Unmarshal(env1.Events[0].Items[0], &m1))

		assert.NotContains(t, fs.Files, fname1)
		assert.Equal(t, 1, len(fs.Files))

		for name := range fs.Files {
			assert.Regexp(t,
				t.Name()+"/sub/f1-([a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}){1}-post.txt", name)
		}
	})
}
