package filegateway

import (
	"bytes"
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"sync"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	xbus "xbus.io/go-xbus/v3"
	"xbus.io/go-xbus/v3/testutils"
)

func writeFile(fs Filesystem, path, content string) error {
	writer, err := fs.Write(path, -1)
	if err != nil {
		return err
	}
	if _, err := writer.Write([]byte(content)); err != nil {
		return err
	}
	return writer.Close()
}

func FilesystemTestSuite(t *testing.T, factory FilesystemFactory, options FilesystemOptions) {
	options.PollingInterval = 10 * time.Millisecond
	t.Run("Write and read", func(t *testing.T) {
		fs, err := factory(options)
		if err != nil {
			t.Fatal(err)
		}
		writer, err := fs.Write(t.Name(), -1)
		if err != nil {
			t.Fatal(err)
		}
		if _, err := writer.Write([]byte("Hello World!")); err != nil {
			t.Fatal(err)
		}
		if err := writer.Close(); err != nil {
			t.Fatal(err)
		}
		reader, err := fs.Read(t.Name())
		if err != nil {
			t.Fatal(err)
		}
		defer func() {
			assert.Nil(t, reader.Close())
		}()
		b, err := ioutil.ReadAll(reader)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, "Hello World!", string(b))
	})

	t.Run("Delete", func(t *testing.T) {
		fs, err := factory(options)
		if err != nil {
			t.Fatal(err)
		}
		if err := fs.Delete(t.Name()); !os.IsNotExist(err) {
			t.Fatal("Expected os.ErrNotExist, got", err)
		}
		writer, err := fs.Write(t.Name(), -1)
		if err != nil {
			t.Fatal(err)
		}
		if _, err := writer.Write([]byte("Hello World!")); err != nil {
			t.Fatal(err)
		}
		if err := writer.Close(); err != nil {
			t.Fatal(err)
		}
		if err := fs.Delete(t.Name()); err != nil {
			t.Fatal(err)
		}
		_, err = fs.Read(t.Name())
		if err == nil {
			t.Fatal("Expects a ErrNotExist, got nil")
		}
		assert.Equal(t, true, os.IsNotExist(err), err.Error())
	})

	t.Run("Rename", func(t *testing.T) {
		fs, err := factory(options)
		if err != nil {
			t.Fatal(err)
		}
		if err := fs.Rename(t.Name(), t.Name()+".1"); !os.IsNotExist(err) {
			t.Fatal("Expected os.ErrNotExist, got", err)
		}
		writer, err := fs.Write(t.Name(), -1)
		if err != nil {
			t.Fatal(err)
		}
		if _, err := writer.Write([]byte("Hello World!")); err != nil {
			t.Fatal(err)
		}
		if err := writer.Close(); err != nil {
			t.Fatal(err)
		}
		if err := fs.Rename(t.Name(), t.Name()+".1"); err != nil {
			t.Fatal(err)
		}
		reader, err := fs.Read(t.Name() + ".1")
		if err != nil {
			t.Fatal(err)
		}
		defer reader.Close()
		b, err := ioutil.ReadAll(reader)
		if err != nil {
			t.Fatal(err)
		}
		assert.Equal(t, "Hello World!", string(b))
	})

	t.Run("WaitForDeletion", func(t *testing.T) {
		t.Run("Returns when the file is not here", func(t *testing.T) {
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			assert.Nil(t, fs.WaitForDeletion(context.Background(), t.Name()))
		})
		t.Run("Is cancelable", func(t *testing.T) {
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			writer, err := fs.Write(t.Name(), -1)
			if err != nil {
				t.Fatal(err)
			}
			if _, err := writer.Write([]byte("Hello World!")); err != nil {
				t.Fatal(err)
			}
			if err := writer.Close(); err != nil {
				t.Fatal(err)
			}
			ctx, f := context.WithTimeout(context.Background(), 50*time.Millisecond)
			defer f()
			err = fs.WaitForDeletion(ctx, t.Name())
			if err == nil {
				t.Fatal("Should receive a DeadlineExceeded error, got nil")
			}
			assert.Equal(t, context.DeadlineExceeded, err, err.Error())
		})
		t.Run("Returns when the file is deleted", func(t *testing.T) {
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			assert.Nil(t, fs.WaitForDeletion(context.Background(), t.Name()))
		})
	})

	t.Run("Observe", func(t *testing.T) {
		t.Run("Is cancelable", func(t *testing.T) {
			logger := testutils.GetTestLogger(t)
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			var pattern = t.Name() + "/*.txt"
			ctx, f := context.WithTimeout(context.Background(), 50*time.Millisecond)
			defer f()
			_, err = fs.Observe(ctx, pattern, logger)
			assert.Nil(t, err)
			<-ctx.Done()
		})
		t.Run("Is cancelable when chan is full", func(t *testing.T) {
			logger := testutils.GetTestLogger(t)
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			fname1 := t.Name() + "/f1.txt"
			fname2 := t.Name() + "/f2.txt"
			assert.Nil(t, writeFile(fs, fname1, "f1 content"))
			assert.Nil(t, writeFile(fs, fname2, "f2 content"))
			var pattern = t.Name() + "/*.txt"
			ctx, f := context.WithTimeout(context.Background(), 50*time.Millisecond)
			defer f()

			assert.Nil(t, err)
			_, err = fs.Observe(ctx, pattern, logger)
			assert.Equal(t, nil, err)
			<-ctx.Done()
		})
		t.Run("Returns the files that matches already", func(t *testing.T) {
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			fname1 := t.Name() + "/f1.txt"
			fname2 := t.Name() + "/f2.txt"
			fname3 := t.Name() + "/f3.log"
			assert.Nil(t, writeFile(fs, fname1, "f1 content"))
			assert.Nil(t, writeFile(fs, fname2, "f2 content"))
			assert.Nil(t, writeFile(fs, fname3, "f3 content"))
			var pattern = t.Name() + "/*.txt"

			ctx, cancel := context.WithTimeout(context.Background(), 300*time.Millisecond)
			defer cancel()

			c, err := fs.Observe(ctx, pattern, testutils.GetTestLogger(t))
			if err != nil {
				t.Fatal(err)
			}
			select {
			case fname := <-c:
				assert.Equal(t, fname1, fname)
			case <-ctx.Done():
				t.Fatal("fname1 was never sent")
			}
			select {
			case fname := <-c:
				assert.Equal(t, fname2, fname)
			case <-ctx.Done():
				t.Fatal("fname2 was never sent")
			}
		})
		t.Run("Returns new files when then arrive", func(t *testing.T) {
			fs, err := factory(options)
			if err != nil {
				t.Fatal(err)
			}
			fname1 := t.Name() + "/f1.txt"
			fname2 := t.Name() + "/f2.txt"
			var pattern = t.Name() + "/*.txt"

			ctx, cancel := context.WithTimeout(context.Background(), 500*time.Millisecond)
			defer cancel()

			go func() {
				select {
				case <-ctx.Done():
					return
				case <-time.After(30 * time.Millisecond):
					if err := writeFile(fs, fname1, "f1 content"); err != nil {
						t.Fatal(err)
					}
				}
				select {
				case <-ctx.Done():
					return
				case <-time.After(50 * time.Millisecond):
					if err := writeFile(fs, fname2, "f2 content"); err != nil {
						t.Fatal(err)
					}
				}
			}()

			c, err := fs.Observe(ctx, pattern, testutils.GetTestLogger(t))
			if err != nil {
				t.Fatal(err)
			}

			select {
			case fname := <-c:
				assert.Equal(t, fname1, fname)
			case <-ctx.Done():
				t.Fatal("timeout waiting for fname1")
			}
			select {
			case fname := <-c:
				assert.Equal(t, fname2, fname)
			case <-ctx.Done():
				t.Fatal("timeout waiting for fname2")
			}
		})
	})
}

func TestMemoryFilesystem(t *testing.T) {
	FilesystemTestSuite(t, memoryFilesystemFactory, FilesystemOptions{Host: "."})
}

// Below is a memory implementation of the Filesystem interface, for testing
// purpose

var memfs = make(map[string]*MemoryFilesystem)

type MemoryFilesystem struct {
	host            string
	Files           map[string][]byte
	pollingInterval time.Duration
	fsWriteError    bool
	m               sync.Mutex
}

func NewMemoryFilesystem(host string) *MemoryFilesystem {
	fs := MemoryFilesystem{
		host:            host,
		Files:           make(map[string][]byte),
		pollingInterval: time.Second,
	}
	return &fs
}

func (fs *MemoryFilesystem) String() string {
	return "mock://" + fs.host
}

type memFileWriter struct {
	*bytes.Buffer

	fs   *MemoryFilesystem
	path string
}

func (w *memFileWriter) Close() error {
	w.fs.m.Lock()
	defer w.fs.m.Unlock()

	w.fs.Files[w.path] = w.Buffer.Bytes()
	return nil
}

func (fs *MemoryFilesystem) Write(path string, size int) (io.WriteCloser, error) {
	if fs.fsWriteError {
		return nil, fmt.Errorf("write error")
	}
	return &memFileWriter{
		Buffer: bytes.NewBuffer(nil),
		fs:     fs,
		path:   path,
	}, nil
}

func (fs *MemoryFilesystem) list() (filepaths []string) {
	fs.m.Lock()
	defer fs.m.Unlock()
	for p := range fs.Files {
		filepaths = append(filepaths, p)
	}
	return
}

func (fs *MemoryFilesystem) Observe(
	ctx context.Context, pattern string, logger xbus.Logger,
) (<-chan string, error) {
	c := make(chan string)
	fo := NewFileObserver(ctx, pattern, "", c)
	go func() {
		for {
			fo.AddFiles(fs.list()...)
			select {
			case <-time.After(fs.pollingInterval):
			case <-ctx.Done():
				return
			}
		}
	}()
	return c, nil
}

type FileReader struct {
	*bytes.Buffer
}

func (fr FileReader) Close() error {
	return nil
}

func (fs *MemoryFilesystem) Read(path string) (io.ReadCloser, error) {
	fs.m.Lock()
	defer fs.m.Unlock()

	if buf, ok := fs.Files[path]; ok {
		return FileReader{bytes.NewBuffer(buf)}, nil
	}
	return nil, os.ErrNotExist
}

func (fs *MemoryFilesystem) WaitForDeletion(ctx context.Context, path string) error {
	for {
		fs.m.Lock()
		_, ok := fs.Files[path]
		fs.m.Unlock()
		if !ok {
			return nil
		}
		select {
		case <-ctx.Done():
			return ctx.Err()
		case <-time.After(fs.pollingInterval):
		}
	}
}

func (fs *MemoryFilesystem) Rename(path, newpath string) error {
	fs.m.Lock()
	defer fs.m.Unlock()

	if buf, ok := fs.Files[path]; ok {
		delete(fs.Files, path)
		fs.Files[newpath] = buf
		return nil
	}
	return os.ErrNotExist
}

func (fs *MemoryFilesystem) Delete(path string) error {
	fs.m.Lock()
	defer fs.m.Unlock()

	if _, ok := fs.Files[path]; ok {
		delete(fs.Files, path)
		return nil
	}
	return os.ErrNotExist
}

func getMemoryFilestem(host string) *MemoryFilesystem {
	if fs, ok := memfs[host]; ok {
		return fs
	}
	fs := NewMemoryFilesystem(host)
	memfs[host] = fs
	return fs
}

func memoryFilesystemFactory(options FilesystemOptions) (Filesystem, error) {
	fs := getMemoryFilestem(options.Host)
	fs.pollingInterval = options.PollingInterval
	fserror, err := options.GetCustomStringD("fserror", "")
	if err != nil {
		return nil, err
	}
	fs.fsWriteError = fserror == "onwrite"
	return fs, nil
}

func init() {
	MustRegisterFilesystem("mock", memoryFilesystemFactory)
}
